/*
 * member_engine.c
 *
 * - Maintain the membership information. The structure holds three
 * membership lists (memlists): current memlist (members currently in the
 * group), join memlist (members who have joined the group) and leave
 * memlist (members who have left the group).
 */

#include "seal.h"

static MEMLIST_control_p_t MEMBER_get_list(MEMBER_control_p_t member_engine, 
		int list_type);

/*
 * create the membership group structure for the three memlists
 */
MEMBER_control_p_t MEMBER_init() {
	MEMBER_control_p_t member_engine;

	member_engine = (MEMBER_control_p_t)malloc(sizeof(MEMBER_control_t));

	// malloc the curr list
	member_engine->curr_memlist = (MEMLIST_control_p_t)malloc(sizeof(
				MEMLIST_control_t));
	if (member_engine->curr_memlist == NULL) {
		debug_err("Unable to create the member list\n");
		return NULL;
	}

	// initialize the variables
	member_engine->curr_memlist->head = NULL;
	member_engine->curr_memlist->tail = NULL;
	member_engine->curr_memlist->size = 0;	

	// malloc the join list
	member_engine->join_memlist = (MEMLIST_control_p_t)malloc(sizeof(
				MEMLIST_control_t));
	if (member_engine->join_memlist == NULL) {
		debug_err("Unable to create the member list\n");
		return NULL;
	}

	// initialize the variables
	member_engine->join_memlist->head = NULL;
	member_engine->join_memlist->tail = NULL;
	member_engine->join_memlist->size = 0;	

	// malloc the leave list
	member_engine->leave_memlist = (MEMLIST_control_p_t)malloc(sizeof(
				MEMLIST_control_t));
	if (member_engine->leave_memlist == NULL) {
		debug_err("Unable to create the member list\n");
		return NULL;
	}

	// initialize the variables
	member_engine->leave_memlist->head = NULL;
	member_engine->leave_memlist->tail = NULL;
	member_engine->leave_memlist->size = 0;	

	// initialize the lock
	pthread_mutex_init(&(member_engine->lock), NULL);

	return member_engine;
}

/* 
 * Insert an entry into the corresponding memlist
 */
void MEMBER_insert(MEMBER_control_p_t member_engine, int list_type, 
		char const* guid, int ts) {

	MEMLIST_control_p_t memlist;	
	MEMLIST_entry_p_t new_member_p;

	if ((memlist = MEMBER_get_list(member_engine, list_type)) == NULL) {
		return;
	}

	// malloc and initialize the new member
	new_member_p = (MEMLIST_entry_p_t)malloc(sizeof(MEMLIST_entry_t));
	memset(new_member_p, 0, sizeof(MEMLIST_entry_t));
	memcpy(new_member_p->guid, guid, SEAL_GUID_LEN);
	new_member_p->ts = ts;

	// add the entry according to the order they arrive (insert at the end)
	if (memlist->head != NULL && memlist->tail != NULL) {
		new_member_p->prev = memlist->tail;
		new_member_p->next = NULL;
		memlist->tail->next = new_member_p;
		memlist->tail = new_member_p;
	} else {			// the memlist is empty
		new_member_p->prev = NULL;
		new_member_p->next = NULL;
		memlist->head = new_member_p;
		memlist->tail = new_member_p;
	}

	memlist->size++;		// increment the size
}

/*
 * Delete an entry form the specified memlist
 * Return 1 on success, 0 on error.
 */
int MEMBER_delete(MEMBER_control_p_t member_engine, int list_type, 
		char const* guid) {
	
	MEMLIST_control_p_t memlist;	
	MEMLIST_entry_p_t tmp_member_p;

	if ((memlist = MEMBER_get_list(member_engine, list_type)) == NULL) {
		return 0;
	}
	
	tmp_member_p = memlist->head;
	while (tmp_member_p != NULL) {
		if (memcmp(tmp_member_p->guid, guid, SEAL_GUID_LEN) == 0) {
			if (tmp_member_p->prev != NULL) {
				tmp_member_p->prev->next = tmp_member_p->next;
			} else {    // remove the head
				memlist->head = tmp_member_p->next;
			}
			if (tmp_member_p->next != NULL) {
				tmp_member_p->next->prev = tmp_member_p->prev;
			} else {    // remove the tail
				memlist->tail = tmp_member_p->prev;
			}
			free(tmp_member_p);
			break;
		}
		tmp_member_p = tmp_member_p->next;
	}

	if (tmp_member_p == NULL) {
		return 0;				// cannot delete
	} else {
		memlist->size--;		// decrement the size
	}

	return 1;
}


/*
 * Delete all entries in the specified memlist
 */
void MEMBER_delete_all(MEMBER_control_p_t member_engine, int list_type) {

	MEMLIST_control_p_t memlist;	
	MEMLIST_entry_p_t del_member_p;

	if ((memlist = MEMBER_get_list(member_engine, list_type)) == NULL) {
		return;
	}

	// remove all member items in the list
	while (memlist->head != NULL) {
		del_member_p = memlist->head;
		memlist->head = del_member_p->next;
		free(del_member_p);
	}

	// reset the list
	memlist->head = memlist->tail = NULL;
	memlist->size = 0;
}


/*
 * Return the size of the specified memlist
 * Return -1 on error.
 */
int MEMBER_get_size(MEMBER_control_p_t member_engine, int list_type) {

	MEMLIST_control_p_t memlist;	

	if ((memlist = MEMBER_get_list(member_engine, list_type)) == NULL) {
		return -1;
	}

	return memlist->size;
}


/*
 * Get all guids, and timestamps of the specified memlist, and return the size
 * of the memlist. Return -1 on error.
 */
int MEMBER_get_all(MEMBER_control_p_t member_engine, int list_type,
		char** guids, int* ts) {
	
	MEMLIST_control_p_t memlist;	
	MEMLIST_entry_p_t tmp_member_p;
	int n = 0;			// count number of members

	if ((memlist = MEMBER_get_list(member_engine, list_type)) == NULL) {
		return -1;
	}

	tmp_member_p = memlist->head;
	while (tmp_member_p != NULL) {
		if (guids != NULL) {
			memcpy(guids[n], tmp_member_p->guid, SEAL_GUID_LEN);
		}
		if (ts != NULL) {
			ts[n] = tmp_member_p->ts;
		}
		tmp_member_p = tmp_member_p->next;
		n++;			// increase the counter
	}

	return n;
}

/*
 * Check if the specified memlist is empty. Return 1 if empty, 0 otherwise.
 */
int MEMBER_is_empty(MEMBER_control_p_t member_engine, int list_type) {

	MEMLIST_control_p_t memlist;	
	
	if ((memlist = MEMBER_get_list(member_engine, list_type)) == NULL) {
		return -1;
	}
	
	if (memlist->head == NULL) {
		return 1;
	}
	return 0;
}


/*
 * Print the information of the specified memlist
 */
void MEMBER_print(MEMBER_control_p_t member_engine, int list_type) {

	MEMLIST_control_p_t memlist;	
	MEMLIST_entry_p_t tmp_member_p;
	int n = 0;

	if ((memlist = MEMBER_get_list(member_engine, list_type)) == NULL) {
		return;
	}

	// Print the list
	tmp_member_p = memlist->head;
	while (tmp_member_p != NULL) {
		if (n > 0) {
			printf(" --> ");
		}
		printf("%d) %s", n++, tmp_member_p->guid);
		tmp_member_p = tmp_member_p->next;
	}
	printf("\n");
}

/* 
 * Destroy the member_engine structure
 */
void MEMBER_destroy(MEMBER_control_p_t member_engine) {

	MEMBER_delete_all(member_engine, MEMBER_CURR_LIST);
	MEMBER_delete_all(member_engine, MEMBER_JOIN_LIST);
	MEMBER_delete_all(member_engine, MEMBER_LEAVE_LIST);
	free(member_engine->curr_memlist);
	free(member_engine->join_memlist);
	free(member_engine->leave_memlist);
	free(member_engine);
}

/*---------------------------------------------------------------------
 *				STATIC FUNCTIONS
 *---------------------------------------------------------------------*/
/*
 * Get the memlist
 */
static MEMLIST_control_p_t MEMBER_get_list(MEMBER_control_p_t member_engine, 
		int list_type) {

	MEMLIST_control_p_t memlist = NULL;

	if (list_type == MEMBER_CURR_LIST) {
		memlist = member_engine->curr_memlist;
	} else if (list_type == MEMBER_JOIN_LIST) {
		memlist = member_engine->join_memlist;
	} else if (list_type == MEMBER_LEAVE_LIST) {
		memlist = member_engine->leave_memlist;
	} else {
		debug_err("wrong list type\n");
	}
	
	return memlist;
}
