/*
 * seal.c
 *
 * - SEcure communicAtion Library
 */

#include "seal.h"

/*
 * Initialize the seal structure, and get the basic information for uid,
 * spread_ip, spread_port, and salt. 
 * Return the structure on success
 */
SEAL_session_p_t SEAL_init() {
	
	SEAL_session_p_t seal;

	// malloc the structure
	if ((seal = (SEAL_session_p_t)malloc(sizeof(SEAL_session_t))) == NULL) {
		return NULL;
	}

	// set the basic information
	memset(seal, 0, sizeof(SEAL_session_t));	
	
	// get the uid, spread_ip and spread port from the config file	
	CONFIG_parse_general(seal->uid, seal->spread_ip, &(seal->spread_port));

	// set the salt value	
	seal->salt = 0;

	// set the password
	memset(seal->passwd, 0, SEAL_PASSWD_LEN + 1);

	// set other variables which are changing every session
	memset(seal->group, 0, SEAL_GROUP_LEN);
	seal->mbox = -1;

	seal->receive_thread = NULL;
	seal->process_thread = NULL;
	seal->rekey_send_thread = NULL;
	seal->rekey_poll_thread = NULL;

	seal->keytree_engine = NULL;
	seal->leader_engine = NULL;
	seal->member_engine = NULL;
	seal->msg_queue = NULL;
	seal->packet_engine = NULL;
	seal->pkt_queue = NULL;
	seal->rekey_queue = NULL;
	seal->sesskey_engine = NULL;
#ifdef SIGNATURE
	seal->certkey_engine = NULL;
#endif

	// return the structure	
	return seal;
}

/*
 * Check the correctness of the password, and set the password if correct.
 * Return 1 on success, 0 on password verification failure.
 */
int SEAL_set_passwd(SEAL_session_p_t seal, char* passwd) {

#ifdef SIGNATURE
	FILE* fp;
	char file_name[50];

    // open long-term private key file
	sprintf(file_name, "%s/%s_key.pem", CERTKEY_KEY_DIR, seal->uid);
	if ((fp = fopen(file_name, "r")) == NULL) {
		return 0;
	}

	// dynamic loading of algorithms
	EVP_add_cipher(EVP_des_ede3_cbc());

	// check password
	if (PEM_read_PrivateKey(fp, NULL, NULL, passwd) == NULL) {
		EVP_cleanup();
		fclose(fp);
		return 0;
	}	

	// release cipher algorithm loading and close file
	EVP_cleanup();
	fclose(fp);

	// set the password
	memset(seal->passwd, 0, SEAL_PASSWD_LEN + 1);
	strncpy(seal->passwd, passwd, SEAL_PASSWD_LEN);	
#endif
	return 1;
}


/*
 * Connect to the spread, join the specified group and create all objects.
 * Return 1 on success, 0 on failure
 */
int SEAL_join(SEAL_session_p_t seal, char const* group) {

	char spread_name[100];
	int ret;
	char new_uid[SEAL_UID_LEN + 2]; // new uid string
	int no_try;						// no of tries of connecting spread

#ifdef SIGNATURE                    // signature mode
	char** all_member_uids;         // all possible members in population
	int no_all_member;              // no. of all possible members 
									// (i.e., population size)
	int i;
#endif

	// error checking
	if (seal->mbox >= 0) {			// indicating the session has joined before
		return 0;
	}

#ifdef SIGNATURE
	seal->certkey_engine = CERTKEY_init(seal->uid, seal->passwd);
	no_all_member = CONFIG_parse_group_size(group);
	if (no_all_member <= 0) {
		return 0;
	} else {
		all_member_uids = (char**)malloc(no_all_member * sizeof(char*));
		for (i=0; i<no_all_member; ++i) {
			all_member_uids[i] = (char*)malloc(SEAL_UID_LEN);
		}
		CONFIG_parse_group(group, all_member_uids);
		CERTKEY_pubkey_install(seal->certkey_engine, all_member_uids,
				no_all_member);
	}
#endif

	// initialize the group 
	memcpy(seal->group, group, SEAL_GROUP_LEN);

	// attach the 2-digit to the end of uid. 
	// This ensures the uniqueness of uid to distinguish from previous
	// membership events.
	memset(new_uid, 0, sizeof(new_uid));
	sprintf(new_uid, "%s%02d", seal->uid, (seal->salt++) % 100);

	// connect to the spread daemon
	sprintf(spread_name, "%d@%s", seal->spread_port, seal->spread_ip);
	memset(seal->guid, 0, SEAL_GUID_LEN);
	ret = SP_connect(spread_name, new_uid, 0, 1, &(seal->mbox), seal->guid);
	no_try = 0;
	while (ret < 0 && no_try < 10) {        // connection error
		// poll for connection
		SP_error(ret);
		sleep(1);
		ret = SP_connect(spread_name, new_uid, 0, 1, &(seal->mbox), seal->guid);
		no_try++;
	}

	// error in connecting to spread
	if (ret < 0 && no_try == 10) {
		debug_err("Cannot connect to spread\n");
		return 0;
	}

	// join to a group under the spread daemon
	ret = SP_join(seal->mbox, seal->group);
	no_try = 0;
	while (ret < 0 && no_try < 10) {        // join error
		// poll for join 
		SP_error(ret);
		sleep(1);
		ret = SP_join(seal->mbox, seal->group);
		no_try++;
	}

	// error in join 
	if (ret < 0 && no_try == 10) {
		debug_err("Cannot join to the specified group\n");
		return 0;
	}

	// init the components
	seal->keytree_engine = KEYTREE_init();
	seal->leader_engine = LEADER_init();
	seal->member_engine = MEMBER_init();
	seal->msg_queue = MSGQUEUE_init();
	seal->packet_engine = PACKET_init(seal->mbox, seal->group);
	seal->pkt_queue = PKTQUEUE_init();
	seal->rekey_queue = RQUEUE_init();
	seal->sesskey_engine = SESSKEY_init();

	// init the threads
	seal->receive_thread = (pthread_t*)malloc(sizeof(pthread_t));
	seal->process_thread = (pthread_t*)malloc(sizeof(pthread_t));
	seal->rekey_poll_thread = (pthread_t*)malloc(sizeof(pthread_t));
	seal->rekey_send_thread = (pthread_t*)malloc(sizeof(pthread_t));
	*(seal->rekey_poll_thread) = 0;
	*(seal->rekey_send_thread) = 0;

	// create thread to handle incoming packets 
	pthread_create(seal->receive_thread, NULL, (void*)&RECEIVE_init, seal);
	pthread_create(seal->process_thread, NULL, (void*)&PROCESS_init, seal);

	// success
	return 1;
}


/*
 * Leave the group, disconnect from spread and free all objects.
 * Return 1 on success, 0 on failure.
 */
int SEAL_leave(SEAL_session_p_t seal) {

	// error checking
	if (seal->mbox < 0) {			// indicating the session has left before
		return 0;
	}
    
    debug("Start Destroying..\n");

	// free all possible locks acquired by the main thread in its SEAL
	// functions calling (assumed on fast mutex)
#ifdef SIGNATURE
	pthread_mutex_trylock(&(seal->certkey_engine->lock));
#endif 
	pthread_mutex_trylock(&(seal->keytree_engine->lock));
	pthread_mutex_trylock(&(seal->leader_engine->lock));
	pthread_mutex_trylock(&(seal->member_engine->lock));
	pthread_mutex_trylock(&(seal->msg_queue->lock));
	pthread_mutex_trylock(&(seal->packet_engine->lock));
	pthread_mutex_trylock(&(seal->pkt_queue->lock));
	pthread_mutex_trylock(&(seal->rekey_queue->lock));
	pthread_mutex_trylock(&(seal->sesskey_engine->lock));

	pthread_mutex_unlock(&(seal->sesskey_engine->lock));
	pthread_mutex_unlock(&(seal->rekey_queue->lock));
	pthread_mutex_unlock(&(seal->pkt_queue->lock));
	pthread_mutex_unlock(&(seal->packet_engine->lock));
	pthread_mutex_unlock(&(seal->msg_queue->lock));
	pthread_mutex_unlock(&(seal->member_engine->lock));
	pthread_mutex_unlock(&(seal->leader_engine->lock));
	pthread_mutex_unlock(&(seal->keytree_engine->lock));
#ifdef SIGNATURE
	pthread_mutex_unlock(&(seal->certkey_engine->lock));
#endif 

	// cancel all threads
    if (*(seal->receive_thread) != 0) {
        pthread_cancel(*(seal->receive_thread));
    }
    if (*(seal->process_thread) != 0) {
        pthread_cancel(*(seal->process_thread));
    }
    if (*(seal->rekey_poll_thread) != 0) {
        pthread_cancel(*(seal->rekey_poll_thread));
    }
    if (*(seal->rekey_send_thread) != 0) {
        pthread_cancel(*(seal->rekey_send_thread));
    }

    // wait all threads to cancel
    if (*(seal->receive_thread) != 0) {
        pthread_join(*(seal->receive_thread), NULL);
    }
    if (*(seal->process_thread) != 0) {
        pthread_join(*(seal->process_thread), NULL);
    }
    if (*(seal->rekey_poll_thread) != 0) {
        pthread_join(*(seal->rekey_poll_thread), NULL);
    }
    if (*(seal->rekey_send_thread) != 0) {
        pthread_join(*(seal->rekey_send_thread), NULL);
    }

	// when threads stop, the member can leave the group
	SP_leave(seal->mbox, seal->group);
    SP_disconnect(seal->mbox);

	// free threads
	free(seal->receive_thread);
	free(seal->process_thread);
	free(seal->rekey_poll_thread);
	free(seal->rekey_send_thread);

    // destroy all objects
    KEYTREE_destroy(seal->keytree_engine);
    LEADER_destroy(seal->leader_engine);
    MEMBER_destroy(seal->member_engine);
    MSGQUEUE_destroy(seal->msg_queue);
    PACKET_destroy(seal->packet_engine);
    PKTQUEUE_destroy(seal->pkt_queue);
    RQUEUE_destroy(seal->rekey_queue);
    SESSKEY_destroy(seal->sesskey_engine);

#ifdef SIGNATURE    
    CERTKEY_destroy(seal->certkey_engine);
#endif

	// reset variables which are changing every session
	memset(seal->group, 0, SEAL_GROUP_LEN);
	seal->mbox = -1;

	seal->receive_thread = NULL;
	seal->process_thread = NULL;
	seal->rekey_send_thread = NULL;
	seal->rekey_poll_thread = NULL;

	seal->keytree_engine = NULL;
	seal->leader_engine = NULL;
	seal->member_engine = NULL;
	seal->msg_queue = NULL;
	seal->packet_engine = NULL;
	seal->pkt_queue = NULL;
	seal->rekey_queue = NULL;
	seal->sesskey_engine = NULL;
#ifdef SIGNATURE
	seal->certkey_engine = NULL;
#endif
    
	debug("End Destroying..\n");

    // success
    return 1;
}

/*
 * Retrieve the data from the message queue and decrypt the message data.
 * Return the message length.
 */
int SEAL_recv(SEAL_session_p_t seal, char* sender, void* msg) {
	char plain_data[SEAL_MAX_MSG_SIZE];
	char encrypt_data[SEAL_MAX_MSG_SIZE];		
	char* signed_data;		// used if the signature mode is activated

	int encrypt_len;		// length of the received encrypted data
	int signed_len;			// length of the signed data
	int msg_len;			// length of the message
	
	int rekey_seq;

	int tmp_htonl;

	// error checking
	if (seal->msg_queue == NULL || seal->sesskey_engine == NULL) {
		return -1;
	}
	// dequeue a message
	pthread_mutex_lock(&(seal->msg_queue->lock));
	while (MSGQUEUE_get_size(seal->msg_queue) <= 0) {
		pthread_cond_wait(&(seal->msg_queue->cond), &(seal->msg_queue->lock));
	}
	pthread_mutex_unlock(&(seal->msg_queue->lock));

	// acquire the locks
	pthread_mutex_lock(&(seal->msg_queue->lock));
	pthread_mutex_lock(&(seal->sesskey_engine->lock));
	
	rekey_seq = MSGQUEUE_get_head_rekey_seq(seal->msg_queue);
	encrypt_len = MSGQUEUE_get_head_msg_size(seal->msg_queue);

	// dequeue the data
	memset(encrypt_data, 0, SEAL_MAX_MSG_SIZE);	
	MSGQUEUE_deq(seal->msg_queue, sender, encrypt_data);

	// decrypt the data
	memset(plain_data, 0, SEAL_MAX_MSG_SIZE);
	if (SESSKEY_decrypt(seal->sesskey_engine, encrypt_data, plain_data, 
				encrypt_len, rekey_seq) < 0) {
		msg_len = -1;
	}

	// get the message length	
	memcpy(&msg_len, plain_data, 4);
	msg_len = ntohl(msg_len);

	signed_data = NULL;
	signed_len = -1;
#ifdef SIGNATURE    
	// construct the data that have been signed	
	if (encrypt_len > CERTKEY_LEN) {
		signed_len = msg_len + 8;	// (4 bytes for rekey_seq, 4 for msg_len)
		signed_data = (char*)malloc(signed_len);
		tmp_htonl = htonl(rekey_seq);
		memcpy(signed_data, &tmp_htonl, 4); 
		memcpy(signed_data + 4, plain_data, msg_len + 4);
	}	

    // verify the signature
    pthread_mutex_lock(&(seal->certkey_engine->lock));
    if (encrypt_len > CERTKEY_LEN && CERTKEY_verify(seal->certkey_engine, 
				sender, signed_data, signed_len,
				plain_data + 4 + msg_len, CERTKEY_LEN)) {
        debug("Verification succeeded\n");
    } else {
        debug_err("Verification failed for sender uid: %s\n", sender);
        msg_len = -1;
    }
    pthread_mutex_unlock(&(seal->certkey_engine->lock));

	// free the resources
	free(signed_data);
#endif

	// set the messages
	memcpy(msg, plain_data + 4, msg_len);

	// release the locks
	pthread_mutex_unlock(&(seal->sesskey_engine->lock));
	pthread_mutex_unlock(&(seal->msg_queue->lock));

	return msg_len;
}

/*
 * Encrypt the message data and send it to the network
 * Return the length of data that have sent, -1 if not.
 */
int SEAL_send(SEAL_session_p_t seal, void const* msg, int msg_len) {
	char plain_data[SEAL_MAX_MSG_SIZE];		// including msg_len + msg
	char encrypt_data[SEAL_MAX_MSG_SIZE];
	char* signed_data;				// used if signature mode is activated

	int plain_len;					// length of plain data
	int encrypt_len;				// length of the encrypted data
	int signed_len;					// length of signed data
	
	int rekey_seq;                  // rekey sequence
	
	int tmp_htonl;					// storing network-byte order number
	
	PACKET_msg_t pkt_msg;			// the message packet to be sent
	int send_len;					// length of the sending packet

	
	// error checking
	if (seal->packet_engine == NULL || seal->sesskey_engine == NULL) {
		return -1;
	}

	// acquire the locks
	pthread_mutex_lock(&(seal->packet_engine->lock));
	pthread_mutex_lock(&(seal->sesskey_engine->lock));

	// get the latest rekey sequence number without getting the group key
	rekey_seq = SESSKEY_groupkey_retrieve(seal->sesskey_engine, NULL, 0);
	if (rekey_seq < 0) {
		pthread_mutex_unlock(&(seal->sesskey_engine->lock));
		pthread_mutex_unlock(&(seal->packet_engine->lock));
		return -1;
	}

	// prepare the message (including both msg_len and messages)
	memset(plain_data, 0, SEAL_MAX_MSG_SIZE);
	tmp_htonl = htonl(msg_len);
	memcpy(plain_data, &tmp_htonl, 4);
	memcpy(plain_data + 4, msg, msg_len);
	plain_len = msg_len + 4;

	signed_data = NULL;
	signed_len = -1;	
#ifdef SIGNATURE
	// construct the data to be signed
	signed_len = plain_len + 4;
	signed_data = (char*)malloc(signed_len);
	tmp_htonl = htonl(rekey_seq);
	memcpy(signed_data, &tmp_htonl, 4); 
	memcpy(signed_data + 4, plain_data, plain_len);

	// sign the data
	pthread_mutex_lock(&(seal->certkey_engine->lock));
	if (CERTKEY_sign(seal->certkey_engine, signed_data, signed_len,
				plain_data + plain_len) == 0) {
		debug_err("Signature failed\n");
	}
	pthread_mutex_unlock(&(seal->certkey_engine->lock));

	// set the new plain data len
	plain_len += CERTKEY_LEN;

	// free the resources
	free(signed_data);
#endif

	// set the encrypt length
	encrypt_len = (plain_len / 8 + 1) * 8;

	// encrypt data
	memset(encrypt_data, 0, SEAL_MAX_MSG_SIZE);
	SESSKEY_encrypt(seal->sesskey_engine, plain_data, encrypt_data,
			encrypt_len);

	// send message packet
	memset(&pkt_msg, 0, sizeof(PACKET_msg_t));
	send_len = 	PACKET_create_msg(rekey_seq, encrypt_data, encrypt_len,
			&pkt_msg);
	PACKET_send(seal->packet_engine, PACKET_MSG, send_len, &pkt_msg);

	// free the locks
	pthread_mutex_unlock(&(seal->sesskey_engine->lock));
	pthread_mutex_unlock(&(seal->packet_engine->lock));

	return send_len;
}

/*
 * Read the membership. Return the guid of the group members.
 */
void SEAL_read_membership(SEAL_session_p_t seal, char** curr_guids, 
		int* no_curr, char** join_guids, int* no_join, char** leave_guids, 
		int* no_leave) {

	// error checking
	if (seal->member_engine == NULL) {
		return;
	}

	pthread_mutex_lock(&(seal->member_engine->lock));

	if (no_curr != NULL) {	
		*no_curr = MEMBER_get_all(seal->member_engine, MEMBER_CURR_LIST, 
				curr_guids, NULL);
	}
	if (no_join != NULL) {
		*no_join = MEMBER_get_all(seal->member_engine, MEMBER_JOIN_LIST,
				join_guids, NULL);
	}
	if (no_leave != NULL) {
		*no_leave = MEMBER_get_all(seal->member_engine, MEMBER_LEAVE_LIST,
				leave_guids, NULL);
	}

	pthread_mutex_unlock(&(seal->member_engine->lock));
}

/*
 * Destroy the seal structure that is created in SEAL_init().
 */
void SEAL_destroy(SEAL_session_p_t seal) {
	free(seal);
}
